#ifndef Algorithm_H
#define Algorithm_H
#include "manager.h"


////////////////////////////////////////////////////////////////////////////////
//
//	Class Name : Algorithm
//
//	Description : 
//		Algorithm class
//
////////////////////////////////////////////////////////////////////////////////
class Algorithm {
private:
  DataManager* data_mgr;
  ALG_TYPE algorithm;

public:
  Algorithm(DataManager* d) { data_mgr = d; }
  DataManager* GetDataMgr(void) const { return data_mgr; }

  ALG_TYPE GetAlg(void) const {return algorithm;}
  void 	   SetAlg(ALG_TYPE alg) {algorithm = alg;}
 
  // Call the algorithm to be run
  void RunAlg(ALG_TYPE alg_type);

  void StartOurAlgorithm(void);
  BOOL StartJoinAlgorithm(Session* sess, UserReq* userRequest);
  BOOL StartLeaveAlgorithm(Session* sess, UserReq* userRequest);
  BOOL ProcessJoinRequest(Session* session, UserReq* userRequest);
  BOOL ProcessLeaveRequest(Session* session, UserReq* userRequest);

  // Dijkstra's algorithm
  Path* ShortestPath(ADJ_TYPE type, int src_node, COST_TYPE cost_type,
			UserReq* userRequest);
  Path* FindCandidateNewPaths(Path* allMinCostPaths, Path* allMinDelayPaths,
					Session* session, UserReq* userRequest);
  Path* FindCandidateNewPaths_SP(Path* allMinCostPaths, Path* allMinDelayPaths,
					Session* session, UserReq* userRequest);
  OnePath* FindCandidateNewPath (OnePath* sharedPath, OnePath* minCostPath,
		OnePath* minDelayPath, Session* session, UserReq* userRequest);
  Path* FindBestS2DPath(Path* candidatePaths, Session* session, UserReq* userRequest);
  BOOL UpdateTreeWithPath(OnePath* shared_onepath, OnePath* new_onepath, Session* sess);
  BOOL UpdateBandwidth(PATH_TYPE pathType, OnePath* sharedPath, Session* session, 
						UserReq* userRequest);

  float JoinCost(OnePath* sharedPath, OnePath* newPath, 
		 Session* session, UserReq* userRequest);
  float DelayCost(OnePath* thePath, UserReq* userRequest);
  float CongCost(PATH_TYPE pathType, OnePath* thePath, 
			Session* session, UserReq* userRequest);
  
  BOOL IsValidQOS(OnePath* shared_onepath, OnePath* new_onepath, 
				Session* sess, UserReq* user_req);
  BOOL CheckBandwidth(PATH_TYPE pathType, OnePath* one_path, 
				Session* sess, UserReq* user_req);
  BOOL UpdatePathBandwidthOnLeave(Session* session, OnePath* backPath, int stopNode);
  BOOL FreeSessionBandwidthOnEdge(int startNode, int endNode, Session* session);
  BOOL AddPathToTreePaths(OnePath* sharedPath, OnePath* newPath, 
				Session* session, UserReq* userRequest);
  OnePath* Combine2Paths(OnePath* path1, OnePath* path2);
  BOOL RemovePathFromTreePaths(int leaveNode, Session* session);
  BOOL CheckRequestValidity(UserReq* userRequest);
  void DisplayCandidateS2DPaths(Path* candidatePaths, Session* session,
					UserReq* userRequest);
};

#endif