///////////////////////////////////////////////////////////////////////////////
//
//	File name : multicast.cc		Last updated: Mar 17, 98
//
//	Description :
//		This is main driver file that calls the DataManager, FileIO 
//		and Algorithm classes.
//
//
//	Authors : Jin-Hwee Park / U.T.
// 	Date	: Jan-27-1998
//
///////////////////////////////////////////////////////////////////////////////
#include "file_io.h"
#include "alg.h"

void main(int argc, char* argv[])
{
	char	sfile[126];
	char	c;

	if (argc != 4) {
		fprintf(stderr, "Usage : %s graph_file req_file algorithm\n", argv[0]);
		fprintf(stderr, "algorithm: ``new'', ``sp'' \n");
		exit(1);
	}

	printf("Graph file            : %s \n", argv[1]);
	printf("Sessions/Requests file: %s \n\n", argv[2]);

	// Create a data_manager object
	DataManager* data_mgr = new DataManager;

	// Read network graph file and session/request file.
	// Create the graph (list of NodeItems), the list of sessions,
	// and the list of request groups.  
	// "Store" these data structures in the allocated data_manager object.
	FileIO* file_mgr = new FileIO(data_mgr, argv[1], argv[2]);

	// Set the parameters.  Assign constants for now.
	// The parameters should also be read from a file in the future.
	// Parameters* parms = new Parameters(0.2, 0.4, 0.4, 0.001);
	// data_mgr->SetParameters(parms);

	// Create an "Algorithm" object; set its attribute "data_mgr"
	// to parm "data_mgr" (the above data_manager object).
	// The algorithm will work on the above data_manager object.
	Algorithm alg(data_mgr);
	if (strcmp(argv[3], "new") == 0)
	  alg.SetAlg(PROPOSED_ALG);
	else if (strcmp(argv[3], "sp") == 0)
	  alg.SetAlg(SHORTEST_PATH);
	else
	{ fprintf(stderr, "\nInvalid algorithm!\n");
	  exit(1);
	}

	// Display the network graph, sessions, requests and parameters
	data_mgr->Display();

	// Run the algorithm.
	// The algorithm will work on the above data_manager object.
	alg.RunAlg(alg.GetAlg());

	// Display outputs
	printf("\n[[Final Result ]]\n");

	// For every session, display the MC tree 
	// (list of NodeItems, each having an adjacency list of edges)
	data_mgr->AllMCTree_Display();

	// For every session, display the paths from mcSource 
	// to the other nodes in the MC tree.
	data_mgr->AllPaths_Display();

	// Deallocate dynamic data structures
	delete data_mgr;
	delete file_mgr;

	fprintf(stderr, "Program terminated normally (%s, %s).\n", argv[1], argv[2]);
	printf("Program terminated normally (%s, %s).\n", argv[1], argv[2]);
	exit(0);
}
