///////////////////////////////////////////////////////////////////////////////
//
//	File name : path.cc			Last updated: Feb 24, 98	
//
//	Description :
//		Implementing methods of classes OnePath and Path
//
//	Authors : Jin-Hwee Park / U.T.
// 	Date 	: Jan-22-1998
//
///////////////////////////////////////////////////////////////////////////////

#include "path.h"

//
// Copy attributes of parm "pn" to current object
//
PathNode::PathNode(PathNode* pn)
{
	SetNodeID(pn->GetNodeID());
//	SetType(pn->GetType());
	SetNext(NIL);
}

//
// Copy attributes of parm "data" to current object
//
OnePath::OnePath(OnePath* data)
{
	SetSrcNode(data->GetSrcNode());
	SetDestNode(data->GetDestNode());
	SetWeight(data->GetWeight());
	SetDelay(data->GetDelay());
	SetHop(data->GetHop());
	SetDestNodeReqBand(data->GetDestNodeReqBand());

	// Copy node_list of "data" to current object
	PathNode* pn = data->GetPathNode();  // pn = node_list of "data"
	SetPathNode(NIL);  	      // current_object.node_list = NIL;
	PathNode* pre_pn = NIL;

	while(pn != NIL) {      // for every node in node_list of "data"
		PathNode* new_pn = new PathNode(pn);  // alloc new entry

		// Add new entry to the tail of current_object.node_list
		if (pre_pn == NIL)
			SetPathNode(new_pn);  // current_object.node_list = new_pn
		else
			pre_pn->SetNext(new_pn); // pre_pn->next = new_pn;
		pre_pn = new_pn;
		pn = pn->GetNext();  // pn = pn->next;
	}
	SetNext(NIL);  // current_object.next = NIL;  (next path = NIL)
}

	
OnePath::~OnePath() 
{
	PathNode* tmp;

	// Free the whole node_list
	while(node_list != NIL) {
		tmp = node_list;
		node_list = node_list->GetNext();  // node_list = node_list->next;
		delete tmp;
	}
}

//
// Add a new PathNode (with node ID "n") to the end of node_list
// This method is not used for the time being.
//  
void 	OnePath::AddNodeToEnd(int n) 
{
	PathNode* p_n = new PathNode(n);  // alloc a new PathNode
					  // init node_id to n
	SetDestNode(n);   // dest_node = n;
	if (node_list == NIL) node_list = p_n;  // first and only entry
	else { 
		PathNode* tmp;
		tmp = node_list;

		// Traverse to the end of node_list
		while(tmp->GetNext() != NIL)   // while(tmp->next != NIL)
			tmp = tmp->GetNext();  // tmp = tmp->next

		// Append p_n to the end of node_list
		tmp->SetNext(p_n);  // tmp->next = p_n;
	}
	
	hops ++;  // increment path length
}

//
// Add a new PathNode (with node ID "n") to the front of node_list
//
// Note: AddNodeToHead need not accumulate the weights/delays of edges added to 
// the path.  For reasons, see alg.cc, method ShortestPath; 
// and session.cc, method FindPathFromSrcToDest (AddNodeToHead is called by
// only these two methods).
//
void 	OnePath::AddNodeToHead(int n)
{

	PathNode* p_n = new PathNode(n);  // alloc a new PathNode
					  // init node_id to n
	if (node_list == NIL) {
		node_list = p_n;  // first and only entry
		SetDestNode(n);   // dest_node = n;
	} else {
		p_n->SetNext(node_list);  // p_n->next = node_list;
		node_list = p_n;
	}
	hops ++;  // increment path length
}

//
// Returns TRUE if node with ID "s_node" is on the path.
// Returns FALSE otherwise.
//
BOOL	OnePath::IsInPath(int s_node)
{
	PathNode* pn;

	pn = GetPathNode();  			// pn = node_list;
	while(pn != NIL) {
		if (pn->GetNodeID() == s_node)  // (pn->node_id == s_node)
			return TRUE;
		pn = pn->GetNext();  		// pn = pn->next;
	}
	return FALSE;
}

//
// Display attributes of current object on screen
//
void	OnePath::Display()
{
	PathNode* tmp;

	printf("SRC[%d], DEST[%d], W[%d], D[%d], H[%d] ==> N[%d]",
	  GetSrcNode(), GetDestNode(), GetWeight(), GetDelay(), GetHop(), GetSrcNode());

	tmp = node_list;   // display the path
	while(tmp != NIL) {
		printf(" -> N[%d]", tmp->GetNodeID());  // tmp->node_id
		tmp = tmp->GetNext();			// tmp = tmp->next;
	}
	printf("\n");
}

//
// Print attributes of current object to a file
//
void	OnePath::Display(FILE* fp)
{
	PathNode* tmp;

	fprintf(fp, "SRC[%d], DEST[%d], W[%d], D[%d], H[%d] ==>  N[%d]", 
	  GetSrcNode(), GetDestNode(), GetWeight(), GetDelay(), GetHop(), GetSrcNode());
	
	tmp = node_list;   // display the path
	while(tmp != NIL) {
		fprintf(fp, " -> N[%d]", tmp->GetNodeID()); // tmp->node_id
		tmp = tmp->GetNext();			// tmp = tmp->next;
	}
	fprintf(stderr, "\n");
}

Path::~Path() 
{
	OnePath* tmp;
	// Free the list of paths
	while(m_path != NIL) {
		tmp = m_path;
		m_path = m_path->GetNext();  // m_path = m_path->next
		delete tmp;  		     // free object pointed to by tmp
	}
}

//
// Add a new path (parm "path") to the front of the list of paths
//
void    Path::AddOnePath(OnePath* path)
{
	OnePath* tmp;

	if (m_path == NIL) m_path = path;
	else {
		path->SetNext(m_path);  // path->next = m_path
		m_path = path;
	}
}

//
// Display attributes of every path in the list of paths
//
void 	Path::Display()
{
	// int count = 0;
	OnePath* tmp;

	tmp = m_path;

	while(tmp != NIL) {
		tmp->Display();	       // display attributes of path "tmp"
		tmp = tmp->GetNext();  // tmp = tmp->next
	}
}


//
// Display attributes of every path in the list of paths
//
void 	Path::Display(FILE *fp)
{
	// int count = 0;
	OnePath* tmp;

	tmp = m_path;

	while(tmp != NIL) {
		tmp->Display(fp);	// display attributes of path "tmp"
		tmp = tmp->GetNext();   // tmp = tmp->next
	}
}

//
// Returns the path whose end node ID is "endNode".
// Returns NIL pointer if no such path is found.
//
OnePath*       Path::FindOnePathEndWith(int endNode)
{
	OnePath*	ret_path;

	ret_path = GetPath();	// ret_path = m_path (head of list of paths)

	while(ret_path != NIL) {
		if (ret_path->GetDestNode() == endNode) // (ret_path->dest_node == endNode)
			return ret_path;
		ret_path = ret_path->GetNext();  // ret_path = ret_path->next
	}
	return NIL;  // no path with end node ID "endNode" is found
}

//
// Reverse the path "one_path" (switching start node and end node).
// Keep the same ccost, delay and path length.
// The new reversed path is stored in another object (other than "one_path").
// (Example: new_reversed_path->GenBackPath(old_path))
//
void	OnePath::GenBackPath(const OnePath* one_path)
{
	SetDestNode(one_path->GetSrcNode()); // switch start node and end node
	SetSrcNode(one_path->GetDestNode());
	SetWeight(one_path->GetWeight());    // keep same ccost
	SetDelay(one_path->GetDelay());      // keep same delay
	SetHop(one_path->GetHop());	     // keep same path length

	node_list = NIL;
	PathNode* pn = one_path->GetPathNode();  // pn = one_path->node_list
	PathNode* new_pn;

	int	cur_node = one_path->GetSrcNode(); // cur_node = one_path->src_node

	// Traverse "one_path->node_list" from front to end,
	// adding each entry to the front of the new (reversed) path.
	//
	while(pn != NIL) {
		new_pn = new PathNode(cur_node);  // alloc a new entry
						  // init node_id to cur_node
		new_pn->SetNext(node_list);  // new_pn->next = node_list
		node_list = new_pn;
		cur_node = pn->GetNodeID();  // cur_node = pn->node_id
		pn = pn->GetNext();	     // pn = pn->next
	}
	// Note: The last entry of "one_path->node_list" is not added to the
	//       reversed path.  It becomes start node of the reversed path.
}

//
// Reverse every path in the list of paths
//
BOOL	Path::ConvertBackToFor(void)
{
	OnePath* one_path = GetPath();  // one_path = m_path (head of list of paths)

	while(one_path != NIL) {   // for every path in the list
		one_path->ConvertBackToFor();    // reverse the path
		one_path = one_path->GetNext();  // one_path = one_path->next
	}
}

//
// Reverse the path (switching start node and end node).
// The new reversed path is stored in the same object.
// (Example: my_path->ConvertBackToFor())
// Returns TRUE if the conversion is successful, and FALSE otherwise.
//
BOOL	OnePath::ConvertBackToFor(void)
{
	PathNode* old_pn = GetPathNode();  // old_pn = node_list
	PathNode* pre_pn = NIL;
	PathNode* new_pathlist = NIL;

	// There is no nodes in this path
	if (old_pn == NIL) return FALSE;

	// Reverse the path (list of PathNodes)
	PathNode* new_pn;
	int	cur_node = GetSrcNode();  // cur_node = src_node
	while(old_pn != NIL) {
		new_pn = new PathNode(cur_node);  // alloc a new entry
						  // init node_id to cur_node
		if (new_pathlist == NIL) {
			new_pathlist = new_pn;    // first and only entry
		} else {
			new_pn->SetNext(new_pathlist);  // new_pn->next = new_pathlist
			new_pathlist = new_pn;
		}
		pre_pn = old_pn;
		cur_node = old_pn->GetNodeID(); // cur_node = old_pn->node_id
		old_pn = old_pn->GetNext();	// old_pn = old_pn->next
		delete pre_pn;
	}
	node_list = new_pathlist;
	// Note: The last entry of old node_list is not added to the
	//       reversed path.  It becomes start node of the reversed path.

	// Switch start node and end node
	int tmp = GetDestNode();    // tmp = dest_node
	SetDestNode(GetSrcNode());  // dest_node = src_node
	SetSrcNode(tmp);	    // src_node = tmp

	return TRUE;  // successful conversion
}


//
// Accumulate the costs of a path (weight, delay, hopCount).
// p_type is either SHARED_PATH or NEW_PATH.
// Returns the accumulated costs.
//
COST	OnePath::FindCost(Graph* graph, PATH_TYPE p_type)
{
        PathNode* pn;
	COST 	ret_cost(0, 0, 0);
	int	s_node;	  // start node of an edge
	int	d_node;	  // end node of an edge

	pn = GetPathNode();     // pn = node_list

	s_node = GetSrcNode();  // s_node = src_node (start node of the path)
	while(pn != NIL) {
		d_node = pn->GetNodeID();  // d_node = pn->node_id (end node)
		Node* nd = graph->FindNode(s_node);
		Edge* edge = graph->FindEdge(nd, d_node);

		if (edge == NIL) {
			fprintf(stderr, "FILE[%s] : LINE[%d] -> Fatal error\n",
				__FILE__, __LINE__);
			fprintf(stderr, "Node[%d] -> Edge[%d]\n",
				s_node, pn->GetNodeID());
			exit(1);
		}

		switch(p_type) {
			case SHARED_PATH:  
			case NEW_PATH:	  
				ret_cost.weight += edge->GetWeight();
				ret_cost.delay += edge->GetDelay();
				ret_cost.hopCount++;
				break;
		}
		s_node = d_node;
		pn = pn->GetNext();  // pn = pn->next
	}
	return ret_cost;
}


//
// This method was not implemented correctly.   It not used anywhere.
//
BOOL	Path::AddWholePath(Path* old_path)
{
	if (old_path == NIL) return TRUE;

	OnePath* one_path = old_path->GetPath();  // one_path = old_path->m_path
	while(one_path != NIL) {
		OnePath* new_onepath = new OnePath(one_path);
		AddOnePath(new_onepath);
		one_path = one_path->GetNext();   // I added this line (UT)
	}
	return TRUE;
}


//
// Returns the path whose start node ID is "startNode".
// Returns NIL pointer if no such path is found.
//
OnePath*       Path::FindOnePathStartWith(int startNodeID)
{
	OnePath*	ret_path;

	ret_path = GetPath();	// ret_path = m_path (head of list of paths)

	while(ret_path != NIL) {
		if (ret_path->GetSrcNode() == startNodeID) 
		    // (ret_path->src_node == startNodeID)
			return ret_path;
		ret_path = ret_path->GetNext();  // ret_path = ret_path->next
	}
	return NIL;  // no path with start node ID "startNodeID" is found
}

//
// Add "secondPath" to the end of current path.
// If "secondPath" is NIL, simply return without modifying the current path.
//
void OnePath::AddPathToEnd(OnePath* secondPath)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++++ AddPathToEnd() \n");
#endif

  PathNode* pnode;
  PathNode* tmp;
  //int destNode;

  if (secondPath == NIL) return;

  // Copy the list of PathNodes of "secondPath" to a new list.
  // "secondList" and "endList" keep track of the head and the 
  // tail of the new list, respectively.
  //
  //destNode = secondPath->GetSrcNode();
  PathNode* secondList = new PathNode(secondPath->GetPathNode()->GetNodeID());
  PathNode* endList    = secondList;
  for (tmp  = secondPath->GetPathNode()->GetNext();
       tmp != NIL;
       tmp  = tmp->GetNext())
  {
    //destNode = tmp->GetNodeID();
    pnode = new PathNode(tmp->GetNodeID());
    endList->SetNext(pnode);		// endList->next = pnode;
    endList = pnode;
  }

  // Append the new list of PathNodes (secondList) to the end of current path.
  //
  if (node_list == NIL)		   	// node_list of current path
    node_list = secondList;
  else
  {
    tmp = node_list;	
    while(tmp->GetNext() != NIL)   	// traverse to the end of node_list
      tmp = tmp->GetNext();
    tmp->SetNext(secondList);  	   	// tmp->next = secondList;
  }

  //SetDestNode(destNode);  		// currentPath->dest_node = destNode;
  SetDestNode(secondPath->GetDestNode());
}

//
// Method SetCosts
//
void OnePath::SetCosts(COST dataCosts)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++++ SetCosts \n");
#endif

  weight_sum = dataCosts.weight;
  delay_sum  = dataCosts.delay;
  hops	     = dataCosts.hopCount;
}

//
// Given "thisNodeID", return the node preceeding "thisNodeID" on the path.
// Return -1 if no preceeding node is found.
//
int OnePath::FindPreceedingNode(int thisNodeID)
{
  if (thisNodeID == GetSrcNode())
    return -1;		// start node of the path does not have a preceeding node

  PathNode* tmp = GetPathNode();
  if (thisNodeID == tmp->GetNodeID())
    return(GetSrcNode()); 

  while (tmp->GetNext() != NIL)
  {
    if (tmp->GetNext()->GetNodeID() == thisNodeID)
      return(tmp->GetNodeID());
    tmp = tmp->GetNext();
  }

  return -1;  	// thisNodeID does not exist in the path
}


//
// Removes path whose end node ID is "endNode".
// Returns FALSE if no such path is found.
// Return TRUE if the removal is successful.
//
BOOL Path::RemovePath(int endNode)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++++ RemovePath() (destination [%d]) \n", endNode);
#endif

  OnePath* prev;
  OnePath* tmp = m_path;		// head of list of paths
  while (tmp != NIL && tmp->GetDestNode() != endNode)
  {
    prev = tmp;
    tmp = tmp->GetNext();		// tmp = tmp->next;
  }
  
  if (tmp == NIL)	// no such path is found
    return FALSE;

  if (tmp == m_path)
    m_path = m_path->GetNext();		// m_path = m_path->next;
  else 
    prev->SetNext(tmp->GetNext());	// prev->next = tmp->next;
  delete tmp;
  return TRUE;
}


////////////////////////////////////////////////////////////////////////////////////
//
// Updates:
// Feb 09: added method FindOnePathStartWith() (UT)
// Feb 12: - added attribute userRequest to class OnePath (UT)
//	   - modified method FindCost()
// Feb 17: added method FindPreceedingNode() to class OnePath (UT)
// Feb 24: added method RemovePath() to class Path (UT)
//
////////////////////////////////////////////////////////////////////////////////////